#' RainFARM Stochastic Precipitation Downscaling
#'
#' @author Jost von Hardenberg, ISAC-CNR, \email{j.vonhardenberg@isac.cnr.it}
#'
#' @description Perform RainFARM downscaling.
#' These functions implement the RainFARM stochastic precipitation downscaling method.
#' Adapted for climate downscaling and including orographic correction.
#' References: D'Onofrio et al. 2014, J of Hydrometeorology 15, 830-843; Rebora et. al 2006, JHM 7, 724
#' The functions use julia_setup from the JuliaCall library, 
#' which takes a long time only the first time it is called on a machine
#' 
#' @param pr the precipitation array to downscal
#' @param lon vector or array of longitudes
#' @param lat vector or array of latitudes
#' @param nf refinement factor for downscaling (the output resolution is increased by this factor)
#' @param slope prescribed or automatic spectral slope (default 0. == automatic)
#' @param kmin first wavenumber for automatic spectral slope
#' @param weights matrix (can be obtained using the \code{rfweights} function (default 1. == none)
#' @param nens number of ensemble members to produce
#' @param fglob logical to conseve global precipitation over the domain (default FALSE)
#' @param fsmooth logical to conserve precipitation with a smnoothing kernel (default TRUE)
#' @param verbose verbose output of the Julia package
#'
#' @return A list containing the fine-scale longitudes, latitudes and the sequence of \code{nens} downscaled fields
#'  
#' @import JuliaCall
#' @examples
#' 
#' nf <- 8   # Choose a downscaling by factor 8
#' nens <- 3 # Number of ensemble members
#'
#' # create a test array with dimension 8x8 and 20 timesteps
#' # or provide your own read from a netcdf file
#'
#' pr <- rnorm(8*8*20)
#' dim(pr) <- c(8,8,20)
#' lon_mat <- seq(10,13.5,0.5) # could also be a 2d matrix
#' lat_mat <- seq(40,43.5,0.5)
#'
#' # Create a test array of weights
#' ww <- array(1., dim = c(8 * nf, 8 * nf))
#' 
#' # ... or create proper weights using an external fine-scale climatology file
#' #     Specify a weightsfn filename if you wish to save the weights
#' 
#' ww <- rfweights("./worldclim.nc", nf, lon=lon_mat, lat=lat_mat, weightsfn="", fsmooth = TRUE)
#' 
#' # downscale using weights (ww=1. means do not use weights)
#' r <- rainfarm(pr, lon_mat, lat_mat, nf, fsmooth = TRUE, fglob = FALSE, weights = ww, nens = 2, verbose = TRUE)
#' lon_f <- r$lon
#' lat_f <- r$lat
#' rd1 <- r$data[[1]]
#' rd2 <- r$data[[2]]
#' @export

CST_RainFARM <- function(data, weights = 1., nf,
                         slope = 0, kmin = 1, nens = 1,
                         fglob = FALSE, fsmooth = FALSE,
                         time_dim = NULL,
                         verbose = FALSE) {

  res <- RainFARM(data$exp, data$lon, data$lat,
           weights, nf, nens,
           slope, kmin,
           fglob, fsmooth,
           time_dim, drop_ens_dim = TRUE,
           verbose)
  n_extra_dims <- length(dim(res$data)) - 3
  extra_dims <- NULL
  if (n_extra_dims > 0) {
    extra_dims <- 1:n_extra_dims + 3
  }
  data$exp <- multiApply:::.aperm2(res$data, c(3, 2, 1, extra_dims))

  data$lon <- res$lon
  data$lat <- res$lat
  data$obs <- NULL

  return(data)
}


RainFARM <- function(data, lon, lat, weights = 1., nf,
                     nens = 1, slope = 0, kmin = 1,
                     fglob = FALSE, fsmooth = TRUE,
                     time_dim = NULL, drop_ens_dim = FALSE,
                     verbose = FALSE) {

  # Check/detect time_dim
  if (is.null(time_dim)) {
    time_dim_names <- c("ftime", "sdate", "time")
    time_dim_num <- which(time_dim_names %in% names(dim(data)))
    if (length(time_dim_num) > 0) {
      time_dim <- time_dim_names[time_dim_num[1]]
    } else {
      stop("Could not automatically detect a target time dimension ",
           "in the provided data in 'data'.")
    }
  }

  # Check availability of JuliaCall
  if (!("JuliaCall" %in% rownames(installed.packages()))) {
    stop("Package 'JuliaCall' needs to be installed in order ",
         "to run RainFARM.")
  }

  # Perform common calls
  JuliaCall::julia_setup(useRCall = FALSE)
  julia_library("RainFARM")

  r <- julia_call("lon_lat_fine", lon, lat, as.integer(nf),
                  need_return = "R")
  lon_f <- r[[1]]
  lat_f <- r[[2]]

  # Repeatedly apply .RainFARM
  result <- Apply(data, c("lon", "lat", time_dim), .RainFARM,
                  weights, nf, nens, slope, kmin,
                  fglob, fsmooth, verbose)$output1

  if (drop_ens_dim && nens == 1) {
    dim(result) <- dim(result)[-which(names(dim(result)) == "member")[1]]
  } else {
    if (drop_ens_dim) {
      stop("Requested 'drop_ens_dim = TRUE' but 'nens > 1'. ",
           "Preserving 'member' dimension.")
    }
  }

  return(list(data = result, lon = lon_f, lat = lat_f))
}


.RainFARM <- function(pr, weights, nf, nens, slope, kmin,
                      fglob, fsmooth, verbose) {

  if (slope == 0) {
    fxp <- julia_call("fft3d", pr, need_return = "R")[[1]]
    sx <- julia_call("fitslopex", fxp, kmin = as.integer(kmin),
                     need_return = "R")
  } else {
    sx <- slope
  }

  result_dims <- c(lon = dim(weights)[1],
                   lat = dim(weights)[2],
                   dim(pr)[3],
                   member = nens)
  names(result_dims)[3] <- names(dim(pr))[3]
  r <- array(dim = result_dims)

  for (i in 1:nens) {
    r[, , , i] <- julia_call("rainfarm", pr, sx, as.integer(nf), weights,
                             fglob = fglob, fsmooth = fsmooth,
                             verbose = verbose, need_return = "R")
  }
  return(r)
}
