#' RainFARM Stochastic Precipitation Downscaling
#'
#' @author Jost von Hardenberg, ISAC-CNR, \email{j.vonhardenberg@isac.cnr.it}
#'
#' @description Compute orographic weights from a fine-scale precipitation climatology file.
#' These functions implement the RainFARM stochastic precipitation downscaling method.
#' Adapted for climate downscaling and including orographic correction.
#' References: D'Onofrio et al. 2014, J of Hydrometeorology 15, 830-843; Rebora et. al 2006, JHM 7, 724
#' The functions use julia_setup from the JuliaCall library, 
#' which takes a long time only the first time it is called on a machine
#' 
#' @param orofile filename of a fine-scale precipitation climatology 
#' @param nf refinement factor for downscaling (the output resolution is increased by this factor)
#' @param lon vector or array of longitudes (can be left out if \code{reffile} is used)
#' @param lat vector or array of latitudes (can be left out if \code{reffile} is used)
#' @param reffile filename of reference file (e.g. the file to downscale). Not needed if \code{lon} and \code{lat} are used
#' @param varname optional name of variable to be read from reffile
#' @param weightsfn optional filename to save the weights in netcdf format
#' @param fsmooth logical to use smooth conservation (default) or large-scale box-average conservation
#'
#' @return The matrix containing the weights
#'  
#' @import JuliaCall, ncdf4
#' @examples
#'  
#' nf <- 8   # Choose a downscaling by factor 8
#' nens <- 3 # Number of ensemble members
#'
#' # create a test array with dimension 8x8 and 20 timesteps
#' # or provide your own read from a netcdf file
#'
#' pr <- rnorm(8*8*20)
#' dim(pr) <- c(8,8,20)
#' lon_mat <- seq(10,13.5,0.5) # could also be a 2d matrix
#' lat_mat <- seq(40,43.5,0.5)
#'
#' # Create a test array of weights
#' ww <- array(1., dim = c(8 * nf, 8 * nf))
#' 
#' # ... or create proper weights using an external fine-scale climatology file
#' #     Specify a weightsfn filename if you wish to save the weights
#' 
#' ww <- rfweights("./worldclim.nc", nf, lon=lon_mat, lat=lat_mat, weightsfn="", fsmooth = TRUE)
#' 
#' # downscale using weights (ww=1. means do not use weights)
#' r <- rainfarm(pr, lon_mat, lat_mat, nf, fsmooth = TRUE, fglob = FALSE, weights = ww, nens = 2, verbose = TRUE)
#' lon_f <- r$lon
#' lat_f <- r$lat
#' rd1 <- r$data[[1]]
#' rd2 <- r$data[[2]]
#' @export

rfweights <- function(orofile, nf, lon=NULL, lat=NULL, reffile="",
                      weightsfn="", varname="", fsmooth=TRUE) {

#  library(JuliaCall)
#  library(ncdf4)
  julia_setup(useRCall = FALSE)
  julia_library("RainFARM")
  if ( (reffile == "") & ( (typeof(lon) == "NULL") | (typeof(lat) == "NULL"))) {
     stop("Neither a reference filename nor lon and lat specified!\n")
  }
  if (reffile == "") {
     xvar <- ncdim_def("lon", "degrees_east", lon, longname = "longitude")
     yvar <- ncdim_def("lat", "degrees_north", lat, longname = "latitude")
     tvar <- ncdim_def("time", "days since 01-01-2000", 0., longname = "time",
                       unlim = T)
     my_ncdf <- ncvar_def("grid", "m", list(xvar, yvar, tvar), -999,
                          longname = "grid", prec = "single")
     reffile <- "__reffile.nc"
     ncfile <- nc_create(reffile, my_ncdf, force_v4 = FALSE)
     nc_close(ncfile)
  }

  ww <- julia_call("rfweights", orofile, reffile, as.integer(nf),
                   weightsfn = weightsfn, varname = varname, fsmooth = fsmooth)

  system("rm -f __reffile.nc gridrf_2_*.nc")
  return(ww)
}
